//! This library provides a way to access system information such as CPU load, mounted filesystems,
//! network interfaces, etc.

#[cfg_attr(
    any(
        target_os = "freebsd",
        target_os = "openbsd",
        target_os = "netbsd",
        target_os = "dragonfly",
        target_vendor = "apple"
    ),
    macro_use
)]
extern crate lazy_static;

#[cfg(feature = "serde")]
extern crate the_serde as serde;

pub mod data;
pub mod platform;
mod reader_utils;

pub use self::data::*;
pub use self::platform::Platform;
pub use self::platform::PlatformImpl as System;
